# /*##########################################################################
#
# Copyright (c) 2004-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
:mod:`silx.gui.plot.actions.io` provides a set of QAction relative of inputs
and outputs for a :class:`.PlotWidget`.

The following QAction are available:

- :class:`CopyAction`
- :class:`PrintAction`
- :class:`SaveAction`
"""

__authors__ = ["V.A. Sole", "T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "25/09/2020"

from io import BytesIO
import logging
import sys
import os.path
import traceback
import numpy
from fabio.TiffIO import TiffIO
from fabio.edfimage import EdfImage

from silx.gui import qt, printer
from silx.gui.dialog.GroupDialog import GroupDialog
from silx.io.utils import save1D, savespec, NEXUS_HDF5_EXT
from silx.io.nxdata import save_NXdata

from . import PlotAction
from ...utils.image import convertArrayToQImage


_logger = logging.getLogger(__name__)

_NEXUS_HDF5_EXT_STR = " ".join(["*" + ext for ext in NEXUS_HDF5_EXT])


def selectOutputGroup(h5filename):
    """Open a dialog to prompt the user to select a group in
    which to output data.

    :param str h5filename: name of an existing HDF5 file
    :rtype: str
    :return: Name of output group, or None if the dialog was cancelled
    """
    dialog = GroupDialog()
    dialog.addFile(h5filename)
    dialog.setWindowTitle("Select an output group")
    if not dialog.exec():
        return None
    return dialog.getSelectedDataUrl().data_path()


class SaveAction(PlotAction):
    """QAction for saving Plot content.

    It opens a Save as... dialog.

    :param plot: :class:`.PlotWidget` instance on which to operate.
    :param parent: See :class:`QAction`.
    """

    SNAPSHOT_FILTER_SVG = "Plot Snapshot as SVG (*.svg)"
    SNAPSHOT_FILTER_PNG = "Plot Snapshot as PNG (*.png)"

    DEFAULT_ALL_FILTERS = (SNAPSHOT_FILTER_PNG, SNAPSHOT_FILTER_SVG)

    # Dict of curve filters with CSV-like format
    # Using ordered dict to guarantee filters order
    # Note: '%.18e' is numpy.savetxt default format
    CURVE_FILTERS_TXT = dict(
        (
            (
                "Curve as Raw ASCII (*.txt)",
                {"fmt": "%.18e", "delimiter": " ", "header": False},
            ),
            (
                'Curve as ";"-separated CSV (*.csv)',
                {"fmt": "%.18e", "delimiter": ";", "header": True},
            ),
            (
                'Curve as ","-separated CSV (*.csv)',
                {"fmt": "%.18e", "delimiter": ",", "header": True},
            ),
            (
                "Curve as tab-separated CSV (*.csv)",
                {"fmt": "%.18e", "delimiter": "\t", "header": True},
            ),
            (
                "Curve as OMNIC CSV (*.csv)",
                {"fmt": "%.7E", "delimiter": ",", "header": False},
            ),
            (
                "Curve as SpecFile (*.dat)",
                {"fmt": "%.10g", "delimiter": "", "header": False},
            ),
        )
    )

    CURVE_FILTER_NPY = "Curve as NumPy binary file (*.npy)"

    CURVE_FILTER_NXDATA = "Curve as NXdata (%s)" % _NEXUS_HDF5_EXT_STR

    DEFAULT_CURVE_FILTERS = list(CURVE_FILTERS_TXT.keys()) + [
        CURVE_FILTER_NPY,
        CURVE_FILTER_NXDATA,
    ]

    DEFAULT_ALL_CURVES_FILTERS = ("All curves as SpecFile (*.dat)",)

    IMAGE_FILTER_EDF = "Image data as EDF (*.edf)"
    IMAGE_FILTER_TIFF = "Image data as TIFF (*.tif)"
    IMAGE_FILTER_NUMPY = "Image data as NumPy binary file (*.npy)"
    IMAGE_FILTER_ASCII = "Image data as ASCII (*.dat)"
    IMAGE_FILTER_CSV_COMMA = "Image data as ,-separated CSV (*.csv)"
    IMAGE_FILTER_CSV_SEMICOLON = "Image data as ;-separated CSV (*.csv)"
    IMAGE_FILTER_CSV_TAB = "Image data as tab-separated CSV (*.csv)"
    IMAGE_FILTER_RGB_PNG = "Image as PNG (*.png)"
    IMAGE_FILTER_NXDATA = "Image as NXdata (%s)" % _NEXUS_HDF5_EXT_STR

    DEFAULT_IMAGE_FILTERS = (
        IMAGE_FILTER_EDF,
        IMAGE_FILTER_TIFF,
        IMAGE_FILTER_NUMPY,
        IMAGE_FILTER_ASCII,
        IMAGE_FILTER_CSV_COMMA,
        IMAGE_FILTER_CSV_SEMICOLON,
        IMAGE_FILTER_CSV_TAB,
        IMAGE_FILTER_RGB_PNG,
        IMAGE_FILTER_NXDATA,
    )

    SCATTER_FILTER_NXDATA = "Scatter as NXdata (%s)" % _NEXUS_HDF5_EXT_STR
    DEFAULT_SCATTER_FILTERS = (SCATTER_FILTER_NXDATA,)

    # filters for which we don't want an "overwrite existing file" warning
    DEFAULT_APPEND_FILTERS = (
        CURVE_FILTER_NXDATA,
        IMAGE_FILTER_NXDATA,
        SCATTER_FILTER_NXDATA,
    )

    def __init__(self, plot, parent=None):
        self._filters = {
            "all": {},
            "curve": {},
            "curves": {},
            "image": {},
            "scatter": {},
        }

        self._appendFilters = list(self.DEFAULT_APPEND_FILTERS)

        # Initialize filters
        for nameFilter in self.DEFAULT_ALL_FILTERS:
            self.setFileFilter(
                dataKind="all", nameFilter=nameFilter, func=self._saveSnapshot
            )

        for nameFilter in self.DEFAULT_CURVE_FILTERS:
            self.setFileFilter(
                dataKind="curve", nameFilter=nameFilter, func=self._saveCurve
            )

        for nameFilter in self.DEFAULT_ALL_CURVES_FILTERS:
            self.setFileFilter(
                dataKind="curves", nameFilter=nameFilter, func=self._saveCurves
            )

        for nameFilter in self.DEFAULT_IMAGE_FILTERS:
            self.setFileFilter(
                dataKind="image", nameFilter=nameFilter, func=self._saveImage
            )

        for nameFilter in self.DEFAULT_SCATTER_FILTERS:
            self.setFileFilter(
                dataKind="scatter", nameFilter=nameFilter, func=self._saveScatter
            )

        super(SaveAction, self).__init__(
            plot,
            icon="document-save",
            text="Save as...",
            tooltip="Save curve/image/plot snapshot dialog",
            triggered=self._actionTriggered,
            checkable=False,
            parent=parent,
        )
        self.setShortcut(qt.QKeySequence.Save)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    @staticmethod
    def _errorMessage(informativeText="", parent=None):
        """Display an error message."""
        # TODO issue with QMessageBox size fixed and too small
        msg = qt.QMessageBox(parent)
        msg.setIcon(qt.QMessageBox.Critical)
        msg.setInformativeText(informativeText + " " + str(sys.exc_info()[1]))
        msg.setDetailedText(traceback.format_exc())
        msg.exec()

    def _saveSnapshot(self, plot, filename, nameFilter):
        """Save a snapshot of the :class:`PlotWindow` widget.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter == self.SNAPSHOT_FILTER_PNG:
            fileFormat = "png"
        elif nameFilter == self.SNAPSHOT_FILTER_SVG:
            fileFormat = "svg"
        else:  # Format not supported
            _logger.error("Saving plot snapshot failed: format not supported")
            return False

        plot.saveGraph(filename, fileFormat=fileFormat)
        return True

    def _getAxesLabels(self, item):
        # If curve has no associated label, get the default from the plot
        xlabel = item.getXLabel() or self.plot.getXAxis().getLabel()
        ylabel = item.getYLabel() or self.plot.getYAxis().getLabel()
        return xlabel, ylabel

    def _get1dData(self, item):
        "provide xdata, [ydata], xlabel, [ylabel] and manages error bars"
        xlabel, ylabel = self._getAxesLabels(item)
        x_data = item.getXData(copy=False)
        y_data = item.getYData(copy=False)
        x_err = item.getXErrorData(copy=False)
        y_err = item.getYErrorData(copy=False)
        labels = [ylabel]
        data = [y_data]

        if x_err is not None:
            if numpy.isscalar(x_err):
                data.append(numpy.zeros_like(y_data) + x_err)
                labels.append(xlabel + "_errors")
            elif x_err.ndim == 1:
                data.append(x_err)
                labels.append(xlabel + "_errors")
            elif x_err.ndim == 2:
                data.append(x_err[0])
                labels.append(xlabel + "_errors_below")
                data.append(x_err[1])
                labels.append(xlabel + "_errors_above")

        if y_err is not None:
            if numpy.isscalar(y_err):
                data.append(numpy.zeros_like(y_data) + y_err)
                labels.append(ylabel + "_errors")
            elif y_err.ndim == 1:
                data.append(y_err)
                labels.append(ylabel + "_errors")
            elif y_err.ndim == 2:
                data.append(y_err[0])
                labels.append(ylabel + "_errors_below")
                data.append(y_err[1])
                labels.append(ylabel + "_errors_above")
        return x_data, data, xlabel, labels

    @staticmethod
    def _selectWriteableOutputGroup(filename, parent):
        if (
            os.path.exists(filename)
            and os.path.isfile(filename)
            and os.access(filename, os.W_OK)
        ):
            entryPath = selectOutputGroup(filename)
            if entryPath is None:
                _logger.info("Save operation cancelled")
                return None
            return entryPath
        elif not os.path.exists(filename):
            # create new entry in new file
            return "/entry"
        else:
            SaveAction._errorMessage("Save failed (file access issue)\n", parent=parent)
            return None

    def _saveCurveAsNXdata(self, curve, filename):
        entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
        if entryPath is None:
            return False

        xlabel, ylabel = self._getAxesLabels(curve)

        return save_NXdata(
            filename,
            nxentry_name=entryPath,
            signal=curve.getYData(copy=False),
            axes=[curve.getXData(copy=False)],
            signal_name="y",
            axes_names=["x"],
            signal_long_name=ylabel,
            axes_long_names=[xlabel],
            signal_errors=curve.getYErrorData(copy=False),
            axes_errors=[curve.getXErrorData(copy=True)],
            title=self.plot.getGraphTitle(),
        )

    def _saveCurve(self, plot, filename, nameFilter):
        """Save a curve from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_CURVE_FILTERS:
            return False

        # Check if a curve is to be saved
        curve = plot.getActiveCurve()
        # before calling _saveCurve, if there is no selected curve, we
        # make sure there is only one curve on the graph
        if curve is None:
            curves = plot.getAllCurves()
            if not curves:
                self._errorMessage("No curve to be saved", parent=self.plot)
                return False
            curve = curves[0]

        if nameFilter in self.CURVE_FILTERS_TXT:
            filter_ = self.CURVE_FILTERS_TXT[nameFilter]
            fmt = filter_["fmt"]
            csvdelim = filter_["delimiter"]
            autoheader = filter_["header"]
        else:
            # .npy or nxdata
            fmt, csvdelim, autoheader = ("", "", False)

        if nameFilter == self.CURVE_FILTER_NXDATA:
            return self._saveCurveAsNXdata(curve, filename)

        xdata, data, xlabel, labels = self._get1dData(curve)

        try:
            save1D(
                filename,
                xdata,
                data,
                xlabel,
                labels,
                fmt=fmt,
                csvdelim=csvdelim,
                autoheader=autoheader,
            )
        except IOError:
            self._errorMessage("Save failed\n", parent=self.plot)
            return False

        return True

    def _saveCurves(self, plot, filename, nameFilter):
        """Save all curves from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_ALL_CURVES_FILTERS:
            return False

        curves = plot.getAllCurves()
        if not curves:
            self._errorMessage("No curves to be saved", parent=self.plot)
            return False

        curve = curves[0]
        scanno = 1
        try:
            xdata, data, xlabel, labels = self._get1dData(curve)

            specfile = savespec(
                filename,
                xdata,
                data,
                xlabel,
                labels,
                fmt="%.7g",
                scan_number=1,
                mode="w",
                write_file_header=True,
                close_file=False,
            )
        except IOError:
            self._errorMessage("Save failed\n", parent=self.plot)
            return False

        for curve in curves[1:]:
            try:
                scanno += 1
                xdata, data, xlabel, labels = self._get1dData(curve)
                specfile = savespec(
                    specfile,
                    xdata,
                    data,
                    xlabel,
                    labels,
                    fmt="%.7g",
                    scan_number=scanno,
                    write_file_header=False,
                    close_file=False,
                )
            except IOError:
                self._errorMessage("Save failed\n", parent=self.plot)
                return False
        specfile.close()

        return True

    def _saveImage(self, plot, filename, nameFilter):
        """Save an image from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_IMAGE_FILTERS:
            return False

        image = plot.getActiveImage()
        if image is None:
            qt.QMessageBox.warning(plot, "No Data", "No image to be saved")
            return False

        data = image.getData(copy=False)

        # TODO Use silx.io for writing files
        if nameFilter == self.IMAGE_FILTER_EDF:
            EdfImage(data=data, header={}).write(filename)
            return True

        elif nameFilter == self.IMAGE_FILTER_TIFF:
            tiffFile = TiffIO(filename, mode="w")
            tiffFile.writeImage(data, software="silx")
            return True

        elif nameFilter == self.IMAGE_FILTER_NUMPY:
            try:
                numpy.save(filename, data)
            except IOError:
                self._errorMessage("Save failed\n", parent=self.plot)
                return False
            return True

        elif nameFilter == self.IMAGE_FILTER_NXDATA:
            entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
            if entryPath is None:
                return False
            xorigin, yorigin = image.getOrigin()
            xscale, yscale = image.getScale()
            xaxis = xorigin + xscale * numpy.arange(data.shape[1])
            yaxis = yorigin + yscale * numpy.arange(data.shape[0])
            xlabel, ylabel = self._getAxesLabels(image)
            interpretation = "image" if len(data.shape) == 2 else "rgba-image"

            return save_NXdata(
                filename,
                nxentry_name=entryPath,
                signal=data,
                axes=[yaxis, xaxis],
                signal_name="image",
                axes_names=["y", "x"],
                axes_long_names=[ylabel, xlabel],
                title=plot.getGraphTitle(),
                interpretation=interpretation,
            )

        elif nameFilter in (
            self.IMAGE_FILTER_ASCII,
            self.IMAGE_FILTER_CSV_COMMA,
            self.IMAGE_FILTER_CSV_SEMICOLON,
            self.IMAGE_FILTER_CSV_TAB,
        ):
            csvdelim, filetype = {
                self.IMAGE_FILTER_ASCII: (" ", "txt"),
                self.IMAGE_FILTER_CSV_COMMA: (",", "csv"),
                self.IMAGE_FILTER_CSV_SEMICOLON: (";", "csv"),
                self.IMAGE_FILTER_CSV_TAB: ("\t", "csv"),
            }[nameFilter]

            height, width = data.shape
            rows, cols = numpy.mgrid[0:height, 0:width]
            try:
                save1D(
                    filename,
                    rows.ravel(),
                    (cols.ravel(), data.ravel()),
                    filetype=filetype,
                    xlabel="row",
                    ylabels=["column", "value"],
                    csvdelim=csvdelim,
                    autoheader=True,
                )

            except IOError:
                self._errorMessage("Save failed\n", parent=self.plot)
                return False
            return True

        elif nameFilter == self.IMAGE_FILTER_RGB_PNG:
            # Get displayed image
            rgbaImage = image.getRgbaImageData(copy=False)
            # Convert RGB QImage
            qimage = convertArrayToQImage(rgbaImage[:, :, :3])

            if qimage.save(filename, "PNG"):
                return True
            else:
                _logger.error("Failed to save image as %s", filename)
                qt.QMessageBox.critical(
                    self.parent(), "Save image as", "Failed to save image"
                )

        return False

    def _saveScatter(self, plot, filename, nameFilter):
        """Save an image from the plot.

        :param str filename: The name of the file to write
        :param str nameFilter: The selected name filter
        :return: False if format is not supported or save failed,
                 True otherwise.
        """
        if nameFilter not in self.DEFAULT_SCATTER_FILTERS:
            return False

        if nameFilter == self.SCATTER_FILTER_NXDATA:
            entryPath = self._selectWriteableOutputGroup(filename, parent=self.plot)
            if entryPath is None:
                return False
            scatter = plot.getScatter()

            x = scatter.getXData(copy=False)
            y = scatter.getYData(copy=False)
            z = scatter.getValueData(copy=False)

            xerror = scatter.getXErrorData(copy=False)
            if isinstance(xerror, float):
                xerror = xerror * numpy.ones(x.shape, dtype=numpy.float32)

            yerror = scatter.getYErrorData(copy=False)
            if isinstance(yerror, float):
                yerror = yerror * numpy.ones(x.shape, dtype=numpy.float32)

            xlabel = plot.getGraphXLabel()
            ylabel = plot.getGraphYLabel()

            return save_NXdata(
                filename,
                nxentry_name=entryPath,
                signal=z,
                axes=[x, y],
                signal_name="values",
                axes_names=["x", "y"],
                axes_long_names=[xlabel, ylabel],
                axes_errors=[xerror, yerror],
                title=plot.getGraphTitle(),
            )

    def setFileFilter(self, dataKind, nameFilter, func, index=None, appendToFile=False):
        """Set a name filter to add/replace a file format support

        :param str dataKind:
            The kind of data for which the provided filter is valid.
            One of: 'all', 'curve', 'curves', 'image', 'scatter'
        :param str nameFilter: The name filter in the QFileDialog.
            See :meth:`QFileDialog.setNameFilters`.
        :param callable func: The function to call to perform saving.
            Expected signature is:
            bool func(PlotWidget plot, str filename, str nameFilter)
        :param bool appendToFile: True to append the data into the selected
            file.
        :param integer index: Index of the filter in the final list (or None)
        """
        assert dataKind in ("all", "curve", "curves", "image", "scatter")

        if appendToFile:
            self._appendFilters.append(nameFilter)

        # first append or replace the new filter to prevent colissions
        self._filters[dataKind][nameFilter] = func
        if index is None:
            # we are already done
            return

        # get the current ordered list of keys
        keyList = list(self._filters[dataKind].keys())

        # deal with negative indices
        if index < 0:
            index = len(keyList) + index
            if index < 0:
                index = 0

        if index >= len(keyList):
            # nothing to be done, already at the end
            txt = "Requested index %d impossible, already at the end" % index
            _logger.info(txt)
            return

        # get the new ordered list
        oldIndex = keyList.index(nameFilter)
        del keyList[oldIndex]
        keyList.insert(index, nameFilter)

        # build the new filters
        newFilters = {}
        for key in keyList:
            newFilters[key] = self._filters[dataKind][key]

        # and update the filters
        self._filters[dataKind] = newFilters
        return

    def getFileFilters(self, dataKind):
        """Returns the nameFilter and associated function for a kind of data.

        :param str dataKind:
            The kind of data for which the provided filter is valid.
            On of: 'all', 'curve', 'curves', 'image', 'scatter'
        :return: {nameFilter: function} associations.
        :rtype: dict
        """
        assert dataKind in ("all", "curve", "curves", "image", "scatter")

        return self._filters[dataKind].copy()

    def _actionTriggered(self, checked=False):
        """Handle save action."""
        # Set-up filters
        filters = {}

        # Add image filters if there is an active image
        if self.plot.getActiveImage() is not None:
            filters.update(self._filters["image"].items())

        # Add curve filters if there is a curve to save
        if self.plot.getActiveCurve() is not None or len(self.plot.getAllCurves()) == 1:
            filters.update(self._filters["curve"].items())
        if len(self.plot.getAllCurves()) >= 1:
            filters.update(self._filters["curves"].items())

        # Add scatter filters if there is a scatter
        # todo: CSV
        if self.plot.getScatter() is not None:
            filters.update(self._filters["scatter"].items())

        filters.update(self._filters["all"].items())

        # Create and run File dialog
        dialog = qt.QFileDialog(self.plot)
        dialog.setOption(qt.QFileDialog.DontUseNativeDialog)
        dialog.setWindowTitle("Output File Selection")
        dialog.setModal(1)
        dialog.setNameFilters(list(filters.keys()))

        dialog.setFileMode(qt.QFileDialog.AnyFile)
        dialog.setAcceptMode(qt.QFileDialog.AcceptSave)

        def onFilterSelection(filt_):
            # disable overwrite confirmation for NXdata types,
            # because we append the data to existing files
            if filt_ in self._appendFilters:
                dialog.setOption(qt.QFileDialog.DontConfirmOverwrite)
            else:
                dialog.setOption(qt.QFileDialog.DontConfirmOverwrite, False)

        dialog.filterSelected.connect(onFilterSelection)

        if not dialog.exec():
            return False

        nameFilter = dialog.selectedNameFilter()
        filename = dialog.selectedFiles()[0]
        dialog.close()

        if "(" in nameFilter and ")" == nameFilter.strip()[-1]:
            # Check for correct file extension
            # Extract file extensions as .something
            extensions = [
                ext[ext.find(".") :]
                for ext in nameFilter[nameFilter.find("(") + 1 : -1].split()
            ]
            for ext in extensions:
                if (
                    len(filename) > len(ext)
                    and filename[-len(ext) :].lower() == ext.lower()
                ):
                    break
            else:  # filename has no extension supported in nameFilter, add one
                if len(extensions) >= 1:
                    filename += extensions[0]

        # Handle save
        func = filters.get(nameFilter, None)
        if func is not None:
            return func(self.plot, filename, nameFilter)
        else:
            _logger.error("Unsupported file filter: %s", nameFilter)
            return False


def _plotAsPNG(plot):
    """Save a :class:`Plot` as PNG and return the payload.

    :param plot: The :class:`Plot` to save
    """
    pngFile = BytesIO()
    plot.saveGraph(pngFile, fileFormat="png")
    pngFile.flush()
    pngFile.seek(0)
    data = pngFile.read()
    pngFile.close()
    return data


class PrintAction(PlotAction):
    """QAction for printing the plot.

    It opens a Print dialog.

    Current implementation print a bitmap of the plot area and not vector
    graphics, so printing quality is not great.

    :param plot: :class:`.PlotWidget` instance on which to operate.
    :param parent: See :class:`QAction`.
    """

    def __init__(self, plot, parent=None):
        super(PrintAction, self).__init__(
            plot,
            icon="document-print",
            text="Print...",
            tooltip="Open print dialog",
            triggered=self.printPlot,
            checkable=False,
            parent=parent,
        )
        self.setShortcut(qt.QKeySequence.Print)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    def getPrinter(self):
        """The QPrinter instance used by the PrintAction.

        :rtype: QPrinter
        """
        return printer.getDefaultPrinter()

    def printPlotAsWidget(self):
        """Open the print dialog and print the plot.

        Use :meth:`QWidget.render` to print the plot

        :return: True if successful
        """
        dialog = qt.QPrintDialog(self.getPrinter(), self.plot)
        dialog.setWindowTitle("Print Plot")
        if not dialog.exec():
            return False

        # Print a snapshot of the plot widget at the top of the page
        widget = self.plot.centralWidget()

        painter = qt.QPainter()
        if not painter.begin(self.getPrinter()):
            return False

        pageRect = self.getPrinter().pageRect(qt.QPrinter.DevicePixel)
        xScale = pageRect.width() / widget.width()
        yScale = pageRect.height() / widget.height()
        scale = min(xScale, yScale)

        painter.translate(pageRect.width() / 2.0, 0.0)
        painter.scale(scale, scale)
        painter.translate(-widget.width() / 2.0, 0.0)
        widget.render(painter)
        painter.end()

        return True

    def printPlot(self):
        """Open the print dialog and print the plot.

        Use :meth:`Plot.saveGraph` to print the plot.

        :return: True if successful
        """
        # Init printer and start printer dialog
        dialog = qt.QPrintDialog(self.getPrinter(), self.plot)
        dialog.setWindowTitle("Print Plot")
        if not dialog.exec():
            return False

        # Save Plot as PNG and make a pixmap from it with default dpi
        pngData = _plotAsPNG(self.plot)

        pixmap = qt.QPixmap()
        pixmap.loadFromData(pngData, "png")

        pageRect = self.getPrinter().pageRect(qt.QPrinter.DevicePixel)
        xScale = pageRect.width() / pixmap.width()
        yScale = pageRect.height() / pixmap.height()
        scale = min(xScale, yScale)

        # Draw pixmap with painter
        painter = qt.QPainter()
        if not painter.begin(self.getPrinter()):
            return False

        painter.drawPixmap(
            0, 0, pixmap.width() * scale, pixmap.height() * scale, pixmap
        )
        painter.end()

        return True


class CopyAction(PlotAction):
    """QAction to copy :class:`.PlotWidget` content to clipboard.

    :param plot: :class:`.PlotWidget` instance on which to operate
    :param parent: See :class:`QAction`
    """

    def __init__(self, plot, parent=None):
        super(CopyAction, self).__init__(
            plot,
            icon="edit-copy",
            text="Copy plot",
            tooltip="Copy a snapshot of the plot into the clipboard",
            triggered=self.copyPlot,
            checkable=False,
            parent=parent,
        )
        self.setShortcut(qt.QKeySequence.Copy)
        self.setShortcutContext(qt.Qt.WidgetShortcut)

    def copyPlot(self):
        """Copy plot content to the clipboard as a bitmap."""
        # Save Plot as PNG and make a QImage from it with default dpi
        pngData = _plotAsPNG(self.plot)
        image = qt.QImage.fromData(pngData, "png")
        qt.QApplication.clipboard().setImage(image)



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

from __future__ import annotations

import typing
import logging
from collections.abc import Iterable
from silx.io.url import DataUrl
from silx.gui import qt
from silx.utils.deprecation import deprecated

_logger = logging.getLogger(__name__)


class UrlList(qt.QListWidget):
    """List of URLs with user selection"""

    sigCurrentUrlChanged = qt.Signal(str)
    """Signal emitted when the active/current URL has changed.

    This signal emits the empty string when there is no longer an active URL.
    """

    sigUrlRemoved = qt.Signal(str)
    """Signal emit when an url is removed from the URL list.

    Provides the url (DataUrl) as a string
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._editable = False
        # are we in 'editable' mode: for now if true then we can remove some items from the list

        # menu to be triggered when in edition from right-click
        self._menu = qt.QMenu()
        self._removeAction = qt.QAction(text="Remove", parent=self)
        self._removeAction.setShortcuts(
            [
                # qt.Qt.Key_Delete,
                qt.QKeySequence.Delete,
            ]
        )
        self._menu.addAction(self._removeAction)

        # connect signal / Slot
        self.currentItemChanged.connect(self._notifyCurrentUrlChanged)

    def setEditable(self, editable: bool):
        """Toggle whether the user can remove some URLs from the list"""
        if editable != self._editable:
            self._editable = editable
            # discusable choice: should we change the selection mode ? No much meaning
            # to be in ExtendedSelection if we are not in editable mode. But does it has more
            # meaning to change the selection mode ?
            if editable:
                self._removeAction.triggered.connect(self._removeSelectedItems)
                self.addAction(self._removeAction)
            else:
                self._removeAction.triggered.disconnect(self._removeSelectedItems)
                self.removeAction(self._removeAction)

    @deprecated(replacement="addUrls", since_version="2.0")
    def setUrls(self, urls: Iterable[DataUrl]) -> None:
        self.addUrls(urls)

    def addUrls(self, urls: Iterable[DataUrl]) -> None:
        """Append multiple DataUrl to the list"""
        self.addItems([url.path() for url in urls])

    def removeUrl(self, url: str):
        """Remove given URL from the list"""
        sel_items = self.findItems(url, qt.Qt.MatchExactly)
        if len(sel_items) > 0:
            assert len(sel_items) == 0, "at most one item expected"
            self.removeItemWidget(sel_items[0])

    def _notifyCurrentUrlChanged(self, current, previous):
        if current is None:
            self.sigCurrentUrlChanged.emit("")
        else:
            self.sigCurrentUrlChanged.emit(current.text())

    def setUrl(self, url: typing.Optional[DataUrl]) -> None:
        """Set the current URL.

        :param url: The new selected URL. Use `None` to clear the selection.
        """
        if url is None:
            self.clearSelection()
            self.sigCurrentUrlChanged.emit("")
        else:
            assert isinstance(url, DataUrl)
            sel_items = self.findItems(url.path(), qt.Qt.MatchExactly)
            if sel_items is None:
                _logger.warning(url.path(), " is not registered in the list.")
            elif len(sel_items) > 0:
                item = sel_items[0]
                self.setCurrentItem(item)
                self.sigCurrentUrlChanged.emit(item.text())

    def _removeSelectedItems(self):
        if not self._editable:
            raise ValueError("UrlList is not set as 'editable'")
        urls = []
        for item in self.selectedItems():
            url = item.text()
            self.takeItem(self.row(item))
            urls.append(url)
        # as the connected slot of 'sigUrlRemoved' can modify the items, better handling all at the end
        for url in urls:
            self.sigUrlRemoved.emit(url)

    def contextMenuEvent(self, event):
        if self._editable:
            globalPos = self.mapToGlobal(event.pos())
            self._menu.exec_(globalPos)



# /*##########################################################################
#
# Copyright (c) 2016-2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Basic tests for CurvesROIWidget"""

__authors__ = ["T. Vincent", "P. Knobel", "H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2017"


import logging
import os.path
import numpy

try:
    from numpy import trapezoid
except ImportError:  # numpy v1 compatibility
    from numpy import trapz as trapezoid

from silx.gui import qt
from silx.gui.plot import items
from silx.gui.plot import Plot1D
from silx.test.utils import temp_dir
from silx.gui.utils.testutils import TestCaseQt, SignalListener
from silx.gui.plot import PlotWindow, CurvesROIWidget
from silx.gui.utils.testutils import getQToolButtonFromAction

_logger = logging.getLogger(__name__)


class TestCurvesROIWidget(TestCaseQt):
    """Basic test for CurvesROIWidget"""

    def setUp(self):
        super(TestCurvesROIWidget, self).setUp()
        self.plot = PlotWindow()
        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

        self.widget = self.plot.getCurvesRoiDockWidget()

        self.widget.show()
        self.qWaitForWindowExposed(self.widget)

    def tearDown(self):
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot

        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        del self.widget

        super(TestCurvesROIWidget, self).tearDown()

    def testDummyAPI(self):
        """Simple test of the getRois and setRois API"""
        roi_neg = CurvesROIWidget.ROI(
            name="negative", fromdata=-20, todata=-10, type_="X"
        )
        roi_pos = CurvesROIWidget.ROI(
            name="positive", fromdata=10, todata=20, type_="X"
        )

        self.widget.roiWidget.setRois((roi_pos, roi_neg))

        rois_defs = self.widget.roiWidget.getRois()
        self.widget.roiWidget.setRois(rois=rois_defs)

    def testWithCurves(self):
        """Plot with curves: test all ROI widget buttons"""
        for offset in range(2):
            self.plot.addCurve(
                numpy.arange(1000),
                offset + numpy.random.random(1000),
                legend=str(offset),
            )

        # Add two ROI
        self.mouseClick(self.widget.roiWidget.addButton, qt.Qt.LeftButton)
        self.qWait(200)
        self.mouseClick(self.widget.roiWidget.addButton, qt.Qt.LeftButton)
        self.qWait(200)

        # Change active curve
        self.plot.setActiveCurve(str(1))

        # Delete a ROI
        self.mouseClick(self.widget.roiWidget.delButton, qt.Qt.LeftButton)
        self.qWait(200)

        with temp_dir() as tmpDir:
            self.tmpFile = os.path.join(tmpDir, "test.ini")

            # Save ROIs
            self.widget.roiWidget.save(self.tmpFile)
            self.assertTrue(os.path.isfile(self.tmpFile))
            self.assertEqual(len(self.widget.getRois()), 2)

            # Reset ROIs
            self.mouseClick(self.widget.roiWidget.resetButton, qt.Qt.LeftButton)
            self.qWait(200)
            rois = self.widget.getRois()
            self.assertEqual(len(rois), 1)
            roiID = list(rois.keys())[0]
            self.assertEqual(rois[roiID].getName(), "ICR")

            # Load ROIs
            self.widget.roiWidget.load(self.tmpFile)
            self.assertEqual(len(self.widget.getRois()), 2)

            del self.tmpFile

    def testMiddleMarker(self):
        """Test with middle marker enabled"""
        self.widget.roiWidget.roiTable.setMiddleROIMarkerFlag(True)

        # Add a ROI
        self.mouseClick(self.widget.roiWidget.addButton, qt.Qt.LeftButton)

        for roiID in self.widget.roiWidget.roiTable._markersHandler._roiMarkerHandlers:
            handler = self.widget.roiWidget.roiTable._markersHandler._roiMarkerHandlers[
                roiID
            ]
            assert handler.getMarker("min")
            xleftMarker = handler.getMarker("min").getXPosition()
            xMiddleMarker = handler.getMarker("middle").getXPosition()
            xRightMarker = handler.getMarker("max").getXPosition()
            thValue = xleftMarker + (xRightMarker - xleftMarker) / 2.0
            self.assertAlmostEqual(xMiddleMarker, thValue)

    def testAreaCalculation(self):
        """Test result of area calculation"""
        x = numpy.arange(100.0)
        y = numpy.arange(100.0)

        # Add two curves
        self.plot.addCurve(x, y, legend="positive")
        self.plot.addCurve(-x, y, legend="negative")

        # Make sure there is an active curve and it is the positive one
        self.plot.setActiveCurve("positive")

        # Add two ROIs
        roi_neg = CurvesROIWidget.ROI(
            name="negative", fromdata=-20, todata=-10, type_="X"
        )
        roi_pos = CurvesROIWidget.ROI(
            name="positive", fromdata=10, todata=20, type_="X"
        )

        self.widget.roiWidget.setRois((roi_pos, roi_neg))

        posCurve = self.plot.getCurve("positive")
        negCurve = self.plot.getCurve("negative")

        self.assertEqual(
            roi_pos.computeRawAndNetArea(posCurve),
            (trapezoid(y=[10, 20], x=[10, 20]), 0.0),
        )
        self.assertEqual(roi_pos.computeRawAndNetArea(negCurve), (0.0, 0.0))
        self.assertEqual(roi_neg.computeRawAndNetArea(posCurve), ((0.0), 0.0))
        self.assertEqual(roi_neg.computeRawAndNetArea(negCurve), ((-150.0), 0.0))

    def testCountsCalculation(self):
        """Test result of count calculation"""
        x = numpy.arange(100.0)
        y = numpy.arange(100.0)

        # Add two curves
        self.plot.addCurve(x, y, legend="positive")
        self.plot.addCurve(-x, y, legend="negative")

        # Make sure there is an active curve and it is the positive one
        self.plot.setActiveCurve("positive")

        # Add two ROIs
        roi_neg = CurvesROIWidget.ROI(
            name="negative", fromdata=-20, todata=-10, type_="X"
        )
        roi_pos = CurvesROIWidget.ROI(
            name="positive", fromdata=10, todata=20, type_="X"
        )

        self.widget.roiWidget.setRois((roi_pos, roi_neg))

        posCurve = self.plot.getCurve("positive")
        negCurve = self.plot.getCurve("negative")

        self.assertEqual(
            roi_pos.computeRawAndNetCounts(posCurve), (y[10:21].sum(), 0.0)
        )
        self.assertEqual(roi_pos.computeRawAndNetCounts(negCurve), (0.0, 0.0))
        self.assertEqual(roi_neg.computeRawAndNetCounts(posCurve), ((0.0), 0.0))
        self.assertEqual(
            roi_neg.computeRawAndNetCounts(negCurve), (y[10:21].sum(), 0.0)
        )

    def testDeferedInit(self):
        """Test behavior of the deferedInit"""
        x = numpy.arange(100.0)
        y = numpy.arange(100.0)
        self.plot.addCurve(x=x, y=y, legend="name", replace="True")
        roisDefs = dict(
            [
                ["range1", dict([["from", 20], ["to", 200], ["type", "energy"]])],
                ["range2", dict([["from", 300], ["to", 500], ["type", "energy"]])],
            ]
        )

        roiWidget = self.plot.getCurvesRoiDockWidget().roiWidget
        self.plot.getCurvesRoiDockWidget().setRois(roisDefs)
        self.assertEqual(len(roiWidget.getRois()), len(roisDefs))
        self.plot.getCurvesRoiDockWidget().setVisible(True)
        self.assertEqual(len(roiWidget.getRois()), len(roisDefs))

    def testDictCompatibility(self):
        """Test that ROI api is valid with dict and not information is lost"""
        roiDict = {
            "from": 20,
            "to": 200,
            "type": "energy",
            "comment": "no",
            "name": "myROI",
            "calibration": [1, 2, 3],
        }
        roi = CurvesROIWidget.ROI._fromDict(roiDict)
        self.assertEqual(roi.toDict(), roiDict)

    def testShowAllROI(self):
        """Test the show allROI action"""
        x = numpy.arange(100.0)
        y = numpy.arange(100.0)
        self.plot.addCurve(x=x, y=y, legend="name", replace="True")

        roisDefsDict = {
            "range1": {"from": 20, "to": 200, "type": "energy"},
            "range2": {"from": 300, "to": 500, "type": "energy"},
        }

        roisDefsObj = (
            CurvesROIWidget.ROI(name="range3", fromdata=20, todata=200, type_="energy"),
            CurvesROIWidget.ROI(
                name="range4", fromdata=300, todata=500, type_="energy"
            ),
        )
        self.widget.roiWidget.showAllMarkers(True)
        roiWidget = self.plot.getCurvesRoiDockWidget().roiWidget
        roiWidget.setRois(roisDefsDict)
        markers = [
            item for item in self.plot.getItems() if isinstance(item, items.MarkerBase)
        ]
        self.assertEqual(len(markers), 2 * 3)

        markersHandler = self.widget.roiWidget.roiTable._markersHandler
        roiWidget.showAllMarkers(True)
        ICRROI = markersHandler.getVisibleRois()
        self.assertEqual(len(ICRROI), 2)

        roiWidget.showAllMarkers(False)
        ICRROI = markersHandler.getVisibleRois()
        self.assertEqual(len(ICRROI), 1)

        roiWidget.setRois(roisDefsObj)
        self.qapp.processEvents()
        markers = [
            item for item in self.plot.getItems() if isinstance(item, items.MarkerBase)
        ]
        self.assertEqual(len(markers), 2 * 3)

        markersHandler = self.widget.roiWidget.roiTable._markersHandler
        roiWidget.showAllMarkers(True)
        ICRROI = markersHandler.getVisibleRois()
        self.assertEqual(len(ICRROI), 2)

        roiWidget.showAllMarkers(False)
        ICRROI = markersHandler.getVisibleRois()
        self.assertEqual(len(ICRROI), 1)

    def testRoiEdition(self):
        """Make sure if the ROI object is edited the ROITable will be updated"""
        roi = CurvesROIWidget.ROI(name="linear", fromdata=0, todata=5)
        self.widget.roiWidget.setRois((roi,))

        x = (0, 1, 1, 2, 2, 3)
        y = (1, 1, 2, 2, 1, 1)
        self.plot.addCurve(x=x, y=y, legend="linearCurve")
        self.plot.setActiveCurve(legend="linearCurve")
        self.widget.calculateROIs()

        roiTable = self.widget.roiWidget.roiTable
        indexesColumns = CurvesROIWidget.ROITable.COLUMNS_INDEX
        itemRawCounts = roiTable.item(0, indexesColumns["Raw Counts"])
        itemNetCounts = roiTable.item(0, indexesColumns["Net Counts"])

        self.assertTrue(itemRawCounts.text() == "8.0")
        self.assertTrue(itemNetCounts.text() == "2.0")

        itemRawArea = roiTable.item(0, indexesColumns["Raw Area"])
        itemNetArea = roiTable.item(0, indexesColumns["Net Area"])

        self.assertTrue(itemRawArea.text() == "4.0")
        self.assertTrue(itemNetArea.text() == "1.0")

        roi.setTo(2)
        itemRawArea = roiTable.item(0, indexesColumns["Raw Area"])
        self.assertTrue(itemRawArea.text() == "3.0")
        roi.setFrom(1)
        itemRawArea = roiTable.item(0, indexesColumns["Raw Area"])
        self.assertTrue(itemRawArea.text() == "2.0")

    def testRemoveActiveROI(self):
        """Test widget behavior when removing the active ROI"""
        roi = CurvesROIWidget.ROI(name="linear", fromdata=0, todata=5)
        self.widget.roiWidget.setRois((roi,))

        self.widget.roiWidget.roiTable.setActiveRoi(None)
        self.assertEqual(len(self.widget.roiWidget.roiTable.selectedItems()), 0)
        self.widget.roiWidget.setRois((roi,))
        self.plot.setActiveCurve(legend="linearCurve")
        self.widget.calculateROIs()

    def testEmitCurrentROI(self):
        """Test behavior of the CurvesROIWidget.sigROISignal"""
        roi = CurvesROIWidget.ROI(name="linear", fromdata=0, todata=5)
        self.widget.roiWidget.setRois((roi,))
        signalListener = SignalListener()
        self.widget.roiWidget.sigROISignal.connect(signalListener.partial())
        self.widget.show()
        self.qapp.processEvents()
        self.assertEqual(signalListener.callCount(), 0)
        self.assertIs(self.widget.roiWidget.roiTable.activeRoi, roi)
        roi.setFrom(0.0)
        self.qapp.processEvents()
        self.assertEqual(signalListener.callCount(), 0)
        roi.setFrom(0.3)
        self.qapp.processEvents()
        self.assertEqual(signalListener.callCount(), 1)


class TestRoiWidgetSignals(TestCaseQt):
    """Test Signals emitted by the RoiWidgetSignals"""

    def setUp(self):
        super().setUp()

        self.plot = Plot1D()
        x = range(20)
        y = range(20)
        self.plot.addCurve(x, y, legend="curve0")
        self.listener = SignalListener()
        self.curves_roi_widget = self.plot.getCurvesRoiWidget()
        self.curves_roi_widget.sigROISignal.connect(self.listener)
        assert self.curves_roi_widget.isVisible() is False
        assert self.listener.callCount() == 0
        self.plot.show()
        self.qWaitForWindowExposed(self.plot)

        toolButton = getQToolButtonFromAction(self.plot.getRoiAction())
        self.qapp.processEvents()
        self.mouseClick(widget=toolButton, button=qt.Qt.LeftButton)

        self.curves_roi_widget.show()
        self.qWaitForWindowExposed(self.curves_roi_widget)

    def tearDown(self):
        self.plot.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.plot.close()
        del self.plot

        self.curves_roi_widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.curves_roi_widget.close()
        del self.curves_roi_widget

        super().tearDown()

    def testSigROISignalAddRmRois(self):
        """Test SigROISignal when adding and removing ROIS"""
        self.listener.clear()

        roi1 = CurvesROIWidget.ROI(name="linear", fromdata=0, todata=5)
        self.curves_roi_widget.roiTable.addRoi(roi1)
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.listener.arguments()[0][0]["current"] == "linear")
        self.listener.clear()

        roi2 = CurvesROIWidget.ROI(name="linear2", fromdata=0, todata=5)
        self.curves_roi_widget.roiTable.addRoi(roi2)
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.listener.arguments()[0][0]["current"] == "linear2")
        self.listener.clear()

        self.curves_roi_widget.roiTable.removeROI(roi2)
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.curves_roi_widget.roiTable.activeRoi == roi1)
        self.assertTrue(self.listener.arguments()[0][0]["current"] == "linear")
        self.listener.clear()

        self.curves_roi_widget.roiTable.deleteActiveRoi()
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.curves_roi_widget.roiTable.activeRoi is None)
        self.assertTrue(self.listener.arguments()[0][0]["current"] is None)
        self.listener.clear()

        self.curves_roi_widget.roiTable.addRoi(roi1)
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.listener.arguments()[0][0]["current"] == "linear")
        self.assertTrue(self.curves_roi_widget.roiTable.activeRoi == roi1)
        self.listener.clear()
        self.qapp.processEvents()

        self.curves_roi_widget.roiTable.removeROI(roi1)
        self.qapp.processEvents()
        self.assertEqual(self.listener.callCount(), 1)
        self.assertTrue(self.listener.arguments()[0][0]["current"] == "ICR")
        self.listener.clear()

    def testSigROISignalModifyROI(self):
        """Test SigROISignal when modifying it"""
        self.curves_roi_widget.roiTable.setMiddleROIMarkerFlag(True)
        roi1 = CurvesROIWidget.ROI(name="linear", fromdata=2, todata=5)
        self.curves_roi_widget.roiTable.addRoi(roi1)
        self.curves_roi_widget.roiTable.setActiveRoi(roi1)

        # test modify the roi2 object
        self.listener.clear()
        roi1.setFrom(0.56)
        self.assertEqual(self.listener.callCount(), 1)
        self.listener.clear()
        roi1.setTo(2.56)
        self.assertEqual(self.listener.callCount(), 1)
        self.listener.clear()
        roi1.setName("linear2")
        self.assertEqual(self.listener.callCount(), 1)
        self.listener.clear()
        roi1.setType("new type")
        self.assertEqual(self.listener.callCount(), 1)

        widget = self.plot.getWidgetHandle()
        widget.setFocus(qt.Qt.OtherFocusReason)
        self.plot.raise_()
        self.qapp.processEvents()

        # modify roi limits (from the gui)
        roi_marker_handler = (
            self.curves_roi_widget.roiTable._markersHandler.getMarkerHandler(
                roi1.getID()
            )
        )
        for marker_type in ("min", "max", "middle"):
            with self.subTest(marker_type=marker_type):
                self.listener.clear()
                marker = roi_marker_handler.getMarker(marker_type)
                x_pix, y_pix = self.plot.dataToPixel(
                    marker.getXPosition(), marker.getYPosition()
                )
                self.mouseMove(widget, pos=(x_pix, y_pix))
                self.qWait(100)
                self.mousePress(widget, qt.Qt.LeftButton, pos=(x_pix, y_pix))
                self.mouseMove(widget, pos=(x_pix + 20, y_pix))
                self.qWait(100)
                self.mouseRelease(widget, qt.Qt.LeftButton, pos=(x_pix + 20, y_pix))
                self.qWait(100)
                self.mouseMove(widget, pos=(x_pix, y_pix))
                self.qapp.processEvents()
                self.assertEqual(self.listener.callCount(), 1)

    def testSetActiveCurve(self):
        """Test sigRoiSignal when set an active curve"""
        roi1 = CurvesROIWidget.ROI(name="linear", fromdata=2, todata=5)
        self.curves_roi_widget.roiTable.setActiveRoi(roi1)
        self.listener.clear()
        self.plot.setActiveCurve("curve0")
        self.assertEqual(self.listener.callCount(), 0)



#!/usr/bin/env python
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2022-2023  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""
This module provides a class for CBF byte offset compression/decompression.
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "09/11/2022"
__status__ = "production"


import os
import struct
import numpy
from ..common import ocl, pyopencl, kernel_workgroup_size
from ..processing import BufferDescription, EventDescription, OpenclProcessing

import logging

logger = logging.getLogger(__name__)


class BitshuffleLz4(OpenclProcessing):
    """Perform the bitshuffle-lz4 decompression on the GPU
    See :class:`OpenclProcessing` for optional arguments description.
    :param int cmp_size:
        Size of the raw stream for decompression.
        It can be (slightly) larger than the array.
    :param int dec_size:
        Size of the decompression output array
        (mandatory for decompression)
    :param dtype: dtype of decompressed data
    """

    LZ4_BLOCK_SIZE = 8192

    def __init__(
        self,
        cmp_size,
        dec_size,
        dtype,
        ctx=None,
        devicetype="all",
        platformid=None,
        deviceid=None,
        block_size=None,
        profile=False,
    ):
        """Constructor of the class:

        :param cmp_size: size of the compressed data buffer (in bytes)
        :param dec_size: size of the compressed data buffer (in words)
        :param dtype: data type of one work in decompressed array

        For the other, see the doc of OpenclProcessing
        """
        OpenclProcessing.__init__(
            self,
            ctx=ctx,
            devicetype=devicetype,
            platformid=platformid,
            deviceid=deviceid,
            block_size=block_size,
            profile=profile,
        )
        if self.block_size is None:
            try:
                self.block_size = self.ctx.devices[0].preferred_work_group_size_multiple
            except:
                self.block_size = self.device.max_work_group_size

        self.cmp_size = numpy.uint64(cmp_size)
        self.dec_size = numpy.uint64(dec_size)
        self.dec_dtype = numpy.dtype(dtype)
        self.num_blocks = numpy.uint32(
            (self.dec_dtype.itemsize * self.dec_size + self.LZ4_BLOCK_SIZE - 1)
            // self.LZ4_BLOCK_SIZE
        )

        buffers = [
            BufferDescription("nb_blocks", 1, numpy.uint32, None),
            BufferDescription("block_position", self.num_blocks, numpy.uint64, None),
            BufferDescription("cmp", self.cmp_size, numpy.uint8, None),
            BufferDescription("dec", self.dec_size, self.dec_dtype, None),
        ]

        self.allocate_buffers(buffers, use_array=True)

        self.compile_kernels([os.path.join("codec", "bitshuffle_lz4")])
        self.block_size = min(
            self.block_size,
            kernel_workgroup_size(self.program, "bslz4_decompress_block"),
        )

    def decompress(self, raw, out=None, wg=None, nbytes=None):
        """This function actually performs the decompression by calling the kernels
        :param numpy.ndarray raw: The compressed data as a 1D numpy array of char or string
        :param pyopencl.array out: pyopencl array in which to place the result.
        :param wg: tuneable parameter with the workgroup size.
        :param int nbytes: (Optional) Number of bytes occupied by the chunk in raw.
        :return: The decompressed image as an pyopencl array.
        :rtype: pyopencl.array
        """

        events = []
        with self.sem:
            if nbytes is not None:
                assert nbytes <= raw.size
                len_raw = numpy.uint64(nbytes)
            elif isinstance(raw, pyopencl.Buffer):
                len_raw = numpy.uint64(raw.size)
            else:
                len_raw = numpy.uint64(len(raw))

            if isinstance(raw, pyopencl.array.Array):
                cmp_buffer = raw.data
                num_blocks = self.num_blocks
            elif isinstance(raw, pyopencl.Buffer):
                cmp_buffer = raw
                num_blocks = self.num_blocks
            else:
                if len_raw > self.cmp_size:
                    self.cmp_size = len_raw
                    logger.info("increase cmp buffer size to %s", self.cmp_size)
                    self.cl_mem["cmp"] = pyopencl.array.empty(
                        self.queue, self.cmp_size, dtype=numpy.uint8
                    )
                evt = pyopencl.enqueue_copy(
                    self.queue, self.cl_mem["cmp"].data, raw, is_blocking=False
                )
                events.append(EventDescription("copy raw H -> D", evt))
                cmp_buffer = self.cl_mem["cmp"].data

                dest_size = struct.unpack(">Q", raw[:8])
                self_dest_nbyte = self.dec_size * self.dec_dtype.itemsize
                if dest_size < self_dest_nbyte:
                    num_blocks = numpy.uint32(
                        (dest_size + self.LZ4_BLOCK_SIZE - 1) // self.LZ4_BLOCK_SIZE
                    )
                elif dest_size > self_dest_nbyte:
                    num_blocks = numpy.uint32(
                        (dest_size + self.LZ4_BLOCK_SIZE - 1) // self.LZ4_BLOCK_SIZE
                    )
                    self.cl_mem["dec"] = pyopencl.array.empty(
                        self.queue, dest_size, self.dec_dtype
                    )
                    self.dec_size = dest_size // self.dec_dtype.itemsize
                else:
                    num_blocks = self.num_blocks

            wg = int(wg or self.block_size)

            evt = self.program.lz4_unblock(
                self.queue,
                (1,),
                (1,),
                cmp_buffer,
                len_raw,
                self.cl_mem["block_position"].data,
                num_blocks,
                self.cl_mem["nb_blocks"].data,
            )
            events.append(EventDescription("LZ4 unblock", evt))

            if out is None:
                out = self.cl_mem["dec"]
            else:
                assert out.dtype == self.dec_dtype
                assert out.size == self.dec_size

            evt = self.program.bslz4_decompress_block(
                self.queue,
                (self.num_blocks * wg,),
                (wg,),
                cmp_buffer,
                out.data,
                self.cl_mem["block_position"].data,
                self.cl_mem["nb_blocks"].data,
                numpy.uint8(self.dec_dtype.itemsize),
            )
            events.append(EventDescription("LZ4 decompress", evt))
        self.profile_multi(events)
        return out

    __call__ = decompress



#!/usr/bin/env python
#
#    Project: Bitshuffle-LZ4 decompression in OpenCL
#             https://github.com/silx-kit/silx
#
#    Copyright (C) 2022-2023  European Synchrotron Radiation Facility,
#                             Grenoble, France
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

"""
Test suite for byte-offset decompression 
"""

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "2022 European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "07/11/2022"

import struct
import numpy
import pytest

try:
    import bitshuffle
except ImportError:
    bitshuffle = None
from silx.opencl.common import ocl, pyopencl
from silx.opencl.codec.bitshuffle_lz4 import BitshuffleLz4


TESTCASES = (  # dtype, shape
    ("uint64", (103, 503)),
    ("int64", (101, 509)),
    ("uint32", (229, 659)),
    ("int32", (233, 653)),
    ("uint16", (743, 647)),
    ("int16", (751, 643)),
    ("uint8", (157, 1373)),
    ("int8", (163, 1367)),
)


@pytest.mark.skipif(
    not ocl or not pyopencl or bitshuffle is None,
    reason="PyOpenCl or bitshuffle is missing",
)
class TestBitshuffleLz4:
    """Test pyopencl bishuffle+LZ4 decompression"""

    @staticmethod
    def _create_test_data(shape, lam=100, dtype="uint32"):
        """Create test (image, compressed stream) pair.

        :param shape: Shape of test image
        :param lam: Expectation of interval argument for numpy.random.poisson
        :return: (reference image array, compressed stream)
        """
        ref = numpy.random.poisson(lam, size=shape).astype(dtype)
        raw = (
            struct.pack(">Q", ref.nbytes)
            + b"\x00" * 4
            + bitshuffle.compress_lz4(ref).tobytes()
        )
        return ref, raw

    @pytest.mark.parametrize("dtype,shape", TESTCASES)
    def test_decompress(self, dtype, shape):
        """
        Tests the byte offset decompression on GPU with various configuration
        """
        ref, raw = self._create_test_data(shape=shape, dtype=dtype)
        bs = BitshuffleLz4(len(raw), numpy.prod(shape), dtype=dtype)
        res = bs.decompress(raw).get()
        assert numpy.array_equal(res, ref.ravel()), "Checks decompression works"

    @pytest.mark.parametrize("dtype,shape", TESTCASES)
    def test_decompress_from_buffer(self, dtype, shape):
        """Test reading compressed data from pyopencl Buffer"""
        ref, raw = self._create_test_data(shape=shape, dtype=dtype)

        bs = BitshuffleLz4(0, numpy.prod(shape), dtype=dtype)

        buffer = pyopencl.Buffer(
            bs.ctx,
            flags=pyopencl.mem_flags.COPY_HOST_PTR | pyopencl.mem_flags.READ_ONLY,
            hostbuf=raw,
        )

        res = bs.decompress(buffer).get()
        assert numpy.array_equal(res, ref.ravel()), "Checks decompression works"

    @pytest.mark.parametrize("dtype,shape", TESTCASES)
    def test_decompress_from_array(self, dtype, shape):
        """Test reading compressed data from pyopencl Array"""
        ref, raw = self._create_test_data(shape=shape, dtype=dtype)

        bs = BitshuffleLz4(0, numpy.prod(shape), dtype=dtype)

        array = pyopencl.array.to_device(
            bs.queue,
            numpy.frombuffer(raw, dtype=numpy.uint8),
            array_queue=bs.queue,
        )

        res = bs.decompress(array).get()
        assert numpy.array_equal(res, ref.ravel()), "Checks decompression works"



# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Access project's data and documentation files.

All access to data and documentation files MUST be made through the functions
of this modules to ensure access across different distribution schemes:

- Installing from source or from wheel
- Installing package as a zip
- Linux packaging willing to install data files (and doc files) in
  alternative folders. In this case, this file must be patched.
- Frozen fat binary application using silx (frozen with cx_Freeze or py2app).
  This needs special care for the resource files in the setup:

  - With cx_Freeze, add silx/resources to include_files:

    .. code-block:: python

       import silx.resources
       silx_include_files = (os.path.dirname(silx.resources.__file__),
                             os.path.join('silx', 'resources'))
       setup(...
             options={'build_exe': {'include_files': [silx_include_files]}}
             )

  - With py2app, add silx in the packages list of the py2app options:

    .. code-block:: python

       setup(...
             options={'py2app': {'packages': ['silx']}}
             )
"""
from __future__ import annotations

__authors__ = ["V.A. Sole", "Thomas Vincent", "J. Kieffer"]
__license__ = "MIT"
__date__ = "08/03/2019"


import atexit
import contextlib
import functools
import logging
import os
import sys
from typing import NamedTuple, Optional

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


logger = logging.getLogger(__name__)


# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/data'
_RESOURCES_DIR = None

# For packaging purpose, patch this variable to use an alternative directory
# E.g., replace with _RESOURCES_DIR = '/usr/share/silx/doc'
# Not in use, uncomment when functionality is needed
# _RESOURCES_DOC_DIR = None

# cx_Freeze frozen support
# See http://cx-freeze.readthedocs.io/en/latest/faq.html#using-data-files
if getattr(sys, "frozen", False):
    # Running in a frozen application:
    # We expect resources to be located either in a silx/resources/ dir
    # relative to the executable or within this package.
    _dir = os.path.join(os.path.dirname(sys.executable), "silx", "resources")
    if os.path.isdir(_dir):
        _RESOURCES_DIR = _dir


class _ResourceDirectory(NamedTuple):
    """Store a source of resources"""

    package_name: str
    forced_path: Optional[str] = None


_SILX_DIRECTORY = _ResourceDirectory(package_name=__name__, forced_path=_RESOURCES_DIR)

_RESOURCE_DIRECTORIES = {}
_RESOURCE_DIRECTORIES["silx"] = _SILX_DIRECTORY


def register_resource_directory(
    name: str, package_name: str, forced_path: Optional[str] = None
):
    """Register another resource directory to the available list.

    By default only the directory "silx" is available.

    .. versionadded:: 0.6

    :param name: Name of the resource directory. It is used on the resource
        name to specify the resource directory to use. The resource
        "silx:foo.png" will use the "silx" resource directory.
    :param package_name: Python name of the package containing resources.
        For example "silx.resources".
    :param forced_path: Path containing the resources. If specified
        neither `importlib` nor `package_name` will be used
        For example "silx.resources".
    :raises ValueError: If the resource directory name already exists.
    """
    if name in _RESOURCE_DIRECTORIES:
        raise ValueError("Resource directory name %s already exists" % name)
    resource_directory = _ResourceDirectory(
        package_name=package_name, forced_path=forced_path
    )
    _RESOURCE_DIRECTORIES[name] = resource_directory


def list_dir(resource: str) -> list[str]:
    """List the content of a resource directory.

    Result are not prefixed by the resource name.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param resource: Name of the resource directory to list
    :return: list of name contained in the directory
    """
    resource_directory, resource_name = _get_package_and_resource(resource)

    if resource_directory.forced_path is not None:
        # if set, use this directory
        path = resource_filename(resource)
        return os.listdir(path)

    path = importlib_resources.files(resource_directory.package_name) / resource_name
    return [entry.name for entry in path.iterdir()]


def is_dir(resource: str) -> bool:
    """True is the resource is a resource directory.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param resource: Name of the resource
    """
    path = resource_filename(resource)
    return os.path.isdir(path)


def exists(resource: str) -> bool:
    """True is the resource exists.

    :param resource: Name of the resource
    """
    path = resource_filename(resource)
    return os.path.exists(path)


def _get_package_and_resource(
    resource: str, default_directory: Optional[str] = None
) -> tuple[_ResourceDirectory, str]:
    """
    Return the resource directory class and a cleaned resource name without
    prefix.

    :param resource: Name of the resource with resource prefix.
    :param default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory.
    :raises ValueError: If the resource name uses an unregistred resource
        directory name
    """
    if ":" in resource:
        prefix, resource = resource.split(":", 1)
    else:
        prefix = "silx"
        if default_directory is not None:
            resource = f"{default_directory}/{resource}"
    if prefix not in _RESOURCE_DIRECTORIES:
        raise ValueError("Resource '%s' uses an unregistred prefix", resource)
    resource_directory = _RESOURCE_DIRECTORIES[prefix]
    return resource_directory, resource


def resource_filename(resource: str) -> str:
    """Return filename corresponding to resource.

    The existence of the resource is not checked.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param resource: Resource path relative to resource directory
                     using '/' path separator. It can be either a file or
                     a directory.
    :raises ValueError: If the resource name uses an unregistred resource
        directory name
    :return: Absolute resource path in the file system
    """
    return _resource_filename(resource, default_directory=None)


# Manage resource files life-cycle
_file_manager = contextlib.ExitStack()
atexit.register(_file_manager.close)


@functools.lru_cache(maxsize=None)
def _get_resource_filename(package: str, resource: str) -> str:
    """Returns path to requested resource in package

    :param package: Name of the package in which to look for the resource
    :param resource: Resource path relative to package using '/' path separator
    :return: Abolute resource path in the file system
    """
    # Caching prevents extracting the resource twice
    file_context = importlib_resources.as_file(
        importlib_resources.files(package) / resource
    )
    path = _file_manager.enter_context(file_context)
    return str(path.absolute())


def _resource_filename(resource: str, default_directory: Optional[str] = None) -> str:
    """Return filename corresponding to resource.

    The existence of the resource is not checked.

    The resource name can be prefixed by the name of a resource directory. For
    example "silx:foo.png" identify the resource "foo.png" from the resource
    directory "silx". See also :func:`register_resource_directory`.

    :param resource: Resource path relative to resource directory
                     using '/' path separator. It can be either a file or
                     a directory.
    :param default_directory: If the resource is not prefixed, the resource
        will be searched on this default directory of the silx resource
        directory. It should only be used internally by silx.
    :return: Absolute resource path in the file system
    """
    resource_directory, resource_name = _get_package_and_resource(
        resource, default_directory=default_directory
    )

    if resource_directory.forced_path is not None:
        # if set, use this directory
        base_dir = resource_directory.forced_path
        resource_path = os.path.join(base_dir, *resource_name.split("/"))
        return resource_path

    return _get_resource_filename(resource_directory.package_name, resource_name)


# Expose ExternalResources for compatibility (since silx 0.11)
from ..utils.ExternalResources import ExternalResources



#!/usr/bin/env python3
# /*##########################################################################
#
# Copyright (c) 2015-2024 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["Jérôme Kieffer", "Thomas Vincent"]
__date__ = "07/11/2022"
__license__ = "MIT"

import sys
import os
import platform
import logging
from typing import Optional

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger("silx.setup")

from setuptools import Extension
from setuptools.command.build_ext import build_ext


PROJECT = "silx"
if sys.version_info.major < 3:
    logger.error(PROJECT + " no longer supports Python2")

if "LANG" not in os.environ and sys.platform == "darwin":
    print(
        """WARNING: the LANG environment variable is not defined,
an utf-8 LANG is mandatory to use setup.py, you may face unexpected UnicodeError.
export LANG=en_US.utf-8
export LC_ALL=en_US.utf-8
"""
    )


# ############## #
# Compiler flags #
# ############## #


def parse_env_as_bool(key: str, default: Optional[bool] = None) -> Optional[bool]:
    """Parse `key` env. var. and convert its value to a boolean or None.

    If it cannot parse it or if None, `default` is returned.
    """
    content = os.environ.get(key, "")
    value = content.lower()
    if value in ["1", "true", "yes", "y"]:
        return True
    if value in ["0", "false", "no", "n"]:
        return False
    if value in ["none", ""]:
        return default
    msg = "Env variable '%s' contains '%s'. But a boolean or an empty \
        string was expected. Variable ignored."
    logger.warning(msg, key, content)
    return default


def get_use_openmp_from_env_var() -> bool:
    """Returns whether or not to build with OpenMP"""
    use_openmp = parse_env_as_bool("SILX_WITH_OPENMP", default=True)
    if use_openmp and platform.system() == "Darwin":
        logger.warning("OpenMP support ignored. Your platform does not support it.")
        return False
    return use_openmp


USE_OPENMP = get_use_openmp_from_env_var()
FORCE_CYTHON = parse_env_as_bool("SILX_FORCE_CYTHON", default=False)


class BuildExt(build_ext):
    """Handle extension compilation.

    Environment variables can custom the build of extensions, see the install documentation.

    If building with MSVC, compiler flags are converted from gcc flags.
    """

    COMPILE_ARGS_CONVERTER = {"-fopenmp": "/openmp"}

    LINK_ARGS_CONVERTER = {"-fopenmp": ""}

    description = "Build extensions"

    def patch_extension(self, ext: Extension):
        """Patch an extension according to requested Cython and OpenMP usage."""
        from Cython.Build import cythonize

        patched_exts = cythonize(
            [ext],
            compiler_directives={"embedsignature": True, "language_level": 3},
            force=FORCE_CYTHON,
        )
        ext.sources = patched_exts[0].sources

        # Remove OpenMP flags if OpenMP is disabled
        if not USE_OPENMP:
            ext.extra_compile_args = [
                f for f in ext.extra_compile_args if f != "-fopenmp"
            ]
            ext.extra_link_args = [f for f in ext.extra_link_args if f != "-fopenmp"]

        # Convert flags from gcc to MSVC if required
        if self.compiler.compiler_type == "msvc":
            extra_compile_args = [
                self.COMPILE_ARGS_CONVERTER.get(f, f) for f in ext.extra_compile_args
            ]
            # Avoid empty arg
            ext.extra_compile_args = [arg for arg in extra_compile_args if arg]

            extra_link_args = [
                self.LINK_ARGS_CONVERTER.get(f, f) for f in ext.extra_link_args
            ]
            # Avoid empty arg
            ext.extra_link_args = [arg for arg in extra_link_args if arg]

    def build_extensions(self):
        for ext in self.extensions:
            self.patch_extension(ext)
        build_ext.build_extensions(self)


# ##### #
# setup #
# ##### #


def get_project_configuration():
    """Returns project arguments for setup"""
    install_requires = [
        # for most of the computation
        "numpy",
        # for version parsing
        "packaging",
        # for io support
        "h5py",
        "fabio>=0.9",
    ]
    if sys.version_info < (3, 9):
        install_requires.append("importlib_resources")

    # extras requirements: target 'full' to install all dependencies at once
    full_requires = [
        # opencl
        "pyopencl",
        "Mako",
        # gui
        "qtconsole",
        "matplotlib>=3.1.0",
        "PyOpenGL",
        "python-dateutil",
        "PyQt5",
        # extra
        "hdf5plugin",
        "scipy",
        "Pillow",
    ]

    test_requires = [
        "pytest>=6.0",
        "pytest-xvfb",
        "pytest-mock",
        # Remove bitshuffle until wheels with numpy 2 support are available
        # "bitshuffle",
        "scipy>=1.10",
        "pooch",
    ]

    doc_requires = {
        "nbsphinx",
        "pandoc",
        "pillow",
        "pydata_sphinx_theme",
        "sphinx",
        "sphinx-autodoc-typehints",
        "sphinx-design",
    }

    extras_require = {
        "full": full_requires,
        "doc": doc_requires,
        "test": test_requires,
    }

    # Here for packaging purpose only
    # Setting the SILX_FULL_INSTALL_REQUIRES environment variable
    # put all dependencies as install_requires
    if os.environ.get("SILX_FULL_INSTALL_REQUIRES") is not None:
        install_requires += full_requires

    # Set the SILX_INSTALL_REQUIRES_STRIP env. var. to a comma-separated
    # list of package names to remove them from install_requires
    install_requires_strip = os.environ.get("SILX_INSTALL_REQUIRES_STRIP")
    if install_requires_strip is not None:
        for package_name in install_requires_strip.split(","):
            install_requires.remove(package_name)

    def silx_io_specfile_define_macros():
        # Locale and platform management
        if sys.platform == "win32":
            return [("WIN32", None), ("SPECFILE_POSIX", None)]
        elif os.name.lower().startswith("posix"):
            # the best choice is to have _GNU_SOURCE defined
            # as a compilation flag because that allows the
            # use of strtod_l
            use_gnu_source = os.environ.get("SPECFILE_USE_GNU_SOURCE", "False")
            if use_gnu_source in ("True", "1"):  # 1 was the initially supported value
                return [("_GNU_SOURCE", 1)]
            return [("SPECFILE_POSIX", None)]
        else:
            return []

    ext_modules = [
        # silx.image
        Extension(
            name="silx.image.bilinear",
            sources=["src/silx/image/bilinear.pyx"],
            language="c",
        ),
        Extension(
            name="silx.image.marchingsquares._mergeimpl",
            sources=["src/silx/image/marchingsquares/_mergeimpl.pyx"],
            include_dirs=[
                os.path.join(
                    os.path.dirname(__file__), "src", "silx", "utils", "include"
                )
            ],
            language="c++",
            extra_link_args=["-fopenmp"],
            extra_compile_args=["-fopenmp"],
        ),
        Extension(
            name="silx.image.shapes",
            sources=["src/silx/image/shapes.pyx"],
            language="c",
        ),
        # silx.io
        Extension(
            name="silx.io.specfile",
            sources=[
                "src/silx/io/specfile/src/sfheader.c",
                "src/silx/io/specfile/src/sfinit.c",
                "src/silx/io/specfile/src/sflists.c",
                "src/silx/io/specfile/src/sfdata.c",
                "src/silx/io/specfile/src/sfindex.c",
                "src/silx/io/specfile/src/sflabel.c",
                "src/silx/io/specfile/src/sfmca.c",
                "src/silx/io/specfile/src/sftools.c",
                "src/silx/io/specfile/src/locale_management.c",
                "src/silx/io/specfile.pyx",
            ],
            define_macros=silx_io_specfile_define_macros(),
            include_dirs=["src/silx/io/specfile/include"],
            language="c",
        ),
        # silx.math
        Extension(
            name="silx.math._colormap",
            sources=["src/silx/math/_colormap.pyx"],
            language="c",
            include_dirs=["src/silx/math/include"],
            extra_link_args=["-fopenmp"],
            extra_compile_args=["-fopenmp"],
        ),
        Extension(
            name="silx.math.chistogramnd",
            sources=[
                "src/silx/math/histogramnd/src/histogramnd_c.c",
                "src/silx/math/chistogramnd.pyx",
            ],
            include_dirs=["src/silx/math/histogramnd/include"],
            language="c",
        ),
        Extension(
            name="silx.math.chistogramnd_lut",
            sources=["src/silx/math/chistogramnd_lut.pyx"],
            include_dirs=["src/silx/math/histogramnd/include"],
            language="c",
        ),
        Extension(
            name="silx.math.combo",
            sources=["src/silx/math/combo.pyx"],
            include_dirs=["src/silx/math/include"],
            language="c",
        ),
        Extension(
            name="silx.math.interpolate",
            sources=["src/silx/math/interpolate.pyx"],
            language="c",
            include_dirs=["src/silx/math/include"],
            extra_link_args=["-fopenmp"],
            extra_compile_args=["-fopenmp"],
        ),
        Extension(
            name="silx.math.marchingcubes",
            sources=[
                "src/silx/math/marchingcubes/mc_lut.cpp",
                "src/silx/math/marchingcubes.pyx",
            ],
            include_dirs=["src/silx/math/marchingcubes"],
            language="c++",
        ),
        Extension(
            name="silx.math.medianfilter.medianfilter",
            sources=["src/silx/math/medianfilter/medianfilter.pyx"],
            include_dirs=["src/silx/math/medianfilter/include"],
            language="c++",
            extra_link_args=["-fopenmp"],
            extra_compile_args=["-fopenmp"],
        ),
        # silx.math.fit
        Extension(
            name="silx.math.fit.filters",
            sources=[
                "src/silx/math/fit/filters/src/smoothnd.c",
                "src/silx/math/fit/filters/src/snip1d.c",
                "src/silx/math/fit/filters/src/snip2d.c",
                "src/silx/math/fit/filters/src/snip3d.c",
                "src/silx/math/fit/filters/src/strip.c",
                "src/silx/math/fit/filters.pyx",
            ],
            include_dirs=["src/silx/math/fit/filters/include"],
            language="c",
        ),
        Extension(
            name="silx.math.fit.functions",
            sources=[
                "src/silx/math/fit/functions/src/funs.c",
                "src/silx/math/fit/functions.pyx",
            ],
            include_dirs=["src/silx/math/fit/functions/include"],
            language="c",
        ),
        Extension(
            name="silx.math.fit.peaks",
            sources=[
                "src/silx/math/fit/peaks/src/peaks.c",
                "src/silx/math/fit/peaks.pyx",
            ],
            include_dirs=["src/silx/math/fit/peaks/include"],
            language="c",
        ),
    ]

    return dict(
        install_requires=install_requires,
        extras_require=extras_require,
        cmdclass=dict(build_ext=BuildExt),
        ext_modules=ext_modules,
    )


if __name__ == "__main__":
    from setuptools import setup

    setup(**get_project_configuration())



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2016-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Example to show the use of :class:`~silx.gui.plot.ImageView.ImageView` widget.

It can be used to open an EDF or TIFF file from the shell command line.

To view an image file with the current installed silx library:
``python examples/imageview.py <file to open>``
To get help:
``python examples/imageview.py -h``
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "08/11/2018"

import logging
import numpy
import time
import threading

from silx.gui.utils import concurrent
from silx.gui.plot.ImageView import ImageViewMainWindow
from silx.gui import qt

logging.basicConfig()
logger = logging.getLogger(__name__)


Nx = 150
Ny = 50


class UpdateThread(threading.Thread):
    """Thread updating the image of a :class:`~sil.gui.plot.Plot2D`

    :param plot2d: The Plot2D to update."""

    def __init__(self, imageview):
        self.imageview = imageview
        self.running = False
        self.future_result = None
        super(UpdateThread, self).__init__()

    def createImage(self, x0: float = 0.0, y0: float = 0.0):
        # width of peak
        sigma_x = 0.15
        sigma_y = 0.25
        # x and y positions
        x = numpy.linspace(-1.5, 1.5, Nx)
        y = numpy.linspace(-1.0, 1.0, Ny)
        xv, yv = numpy.meshgrid(x, y)
        signal = numpy.exp(
            -((xv - x0) ** 2 / sigma_x**2 + (yv - y0) ** 2 / sigma_y**2)
        )
        # add noise
        signal += 0.3 * numpy.random.random(size=signal.shape)
        return signal

    def start(self):
        """Start the update thread"""
        self.running = True
        super(UpdateThread, self).start()

    def run(self):
        """Method implementing thread loop that updates the plot

        It produces an image every 10 ms or so, and
        either updates the plot or skip the image
        """
        x0, y0 = 0.0, 0.0

        while self.running:
            time.sleep(0.01)

            signal = self.createImage(x0, y0)
            x0 += 0.05 * (numpy.random.random() - 0.5)
            y0 += 0.05 * (numpy.random.random() - 0.5)

            # If previous frame was not added to the plot yet, skip this one
            if self.future_result is None or self.future_result.done():
                # plot the data asynchronously, and
                # keep a reference to the `future` object
                self.future_result = concurrent.submitToQtMainThread(
                    self.imageview.setImage, signal, resetzoom=False
                )

    def stop(self):
        """Stop the update thread"""
        self.running = False
        self.join(2)


def main(argv=None):
    """Display an image from a file in an :class:`ImageView` widget.

    :param argv: list of command line arguments or None (the default)
                 to use sys.argv.
    :type argv: list of str
    :return: Exit status code
    :rtype: int
    :raises IOError: if no image can be loaded from the file
    """
    import argparse

    import fabio
    from fabio.fabioimage import FabioImage

    # Command-line arguments
    parser = argparse.ArgumentParser(description="Browse the images of an EDF file.")
    parser.add_argument(
        "-o",
        "--origin",
        nargs=2,
        type=float,
        default=(0.0, 0.0),
        help="""Coordinates of the origin of the image: (x, y).
        Default: 0., 0.""",
    )
    parser.add_argument(
        "-s",
        "--scale",
        nargs=2,
        type=float,
        default=(1.0, 1.0),
        help="""Scale factors applied to the image: (sx, sy).
        Default: 1., 1.""",
    )
    parser.add_argument(
        "-l",
        "--log",
        action="store_true",
        help="Use logarithm normalization for colormap, default: Linear.",
    )
    parser.add_argument("filename", nargs="?", help="Filename of the image to open")
    parser.add_argument(
        "--live", action="store_true", help="Live update of a generated image"
    )
    args = parser.parse_args(args=argv)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication([])
    sys.excepthook = qt.exceptionHandler

    mainWindow = ImageViewMainWindow()
    mainWindow.setAttribute(qt.Qt.WA_DeleteOnClose)
    mainWindow.setFocus(qt.Qt.OtherFocusReason)
    mainWindow.setKeepDataAspectRatio(True)

    if args.log:  # Use log normalization by default
        colormap = mainWindow.getDefaultColormap()
        colormap.setNormalization(colormap.LOGARITHM)

    if args.live:
        # Start updating the plot
        updateThread = UpdateThread(mainWindow)
        updateThread.start()
        mainWindow.setImage(updateThread.createImage())
        mainWindow.show()
        try:
            return app.exec()
        finally:
            updateThread.stop()

    # Open/create input image data
    if args.filename:
        image = fabio.open(args.filename)

    else:
        logger.warning("No image file provided, displaying dummy data")
        size = 512
        xx, yy = numpy.ogrid[-size:size, -size:size]
        data = numpy.cos(xx / (size // 5)) + numpy.cos(yy / (size // 5))
        data = numpy.random.poisson(numpy.abs(data))
        image = FabioImage(data)

    mainWindow.setImage(image.data, origin=args.origin, scale=args.scale)

    if image.nframes > 1:
        # Add a toolbar for multi-frame EDF support
        multiFrameToolbar = qt.QToolBar("Multi-frame")
        multiFrameToolbar.addWidget(qt.QLabel("Frame [0-%d]:" % (image.nframes - 1)))

        spinBox = qt.QSpinBox()
        spinBox.setRange(0, image.nframes - 1)

        def updateImage(index):
            mainWindow.setImage(
                image.get_frame(index).data,
                origin=args.origin,
                scale=args.scale,
                reset=False,
            )

        spinBox.valueChanged[int].connect(updateImage)
        multiFrameToolbar.addWidget(spinBox)

        mainWindow.addToolBar(multiFrameToolbar)

    mainWindow.show()
    return app.exec()


if __name__ == "__main__":
    import sys

    sys.exit(main(argv=sys.argv[1:]))



# /*##########################################################################
#
# Copyright (c) 2014-2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides minimalistic text support for OpenGL.
It provides Latin-1 (ISO8859-1) characters for one monospace font at one size.
"""

from __future__ import annotations

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "03/04/2017"


from collections import OrderedDict
import weakref

import numpy

from .... import qt
from ...._glutils import font, gl, Context, Program, Texture
from .GLSupport import mat4Translate
from silx.gui.colors import RGBAColorType


class _Cache:
    """LRU (Least Recent Used) cache.

    :param int maxsize: Maximum number of (key, value) pairs in the cache
    :param callable callback:
       Called when a (key, value) pair is removed from the cache.
       It must take 2 arguments: key and value.
    """

    def __init__(self, maxsize=128, callback=None):
        self._maxsize = int(maxsize)
        self._callback = callback
        self._cache = OrderedDict()  # Needed for popitem(last=False)

    def __contains__(self, item):
        return item in self._cache

    def __getitem__(self, key):
        if key in self._cache:
            # Remove/add key from ordered dict to store last access info
            value = self._cache.pop(key)
            self._cache[key] = value
            return value
        else:
            raise KeyError

    def __setitem__(self, key, value):
        """Add a key, value pair to the cache.

        :param key: The key to set
        :param value: The corresponding value
        """
        if key not in self._cache and len(self._cache) >= self._maxsize:
            removedKey, removedValue = self._cache.popitem(last=False)
            if self._callback is not None:
                self._callback(removedKey, removedValue)
        self._cache[key] = value


# Text2D ######################################################################

LEFT, CENTER, RIGHT = "left", "center", "right"
TOP, BASELINE, BOTTOM = "top", "baseline", "bottom"
ROTATE_90, ROTATE_180, ROTATE_270 = 90, 180, 270


class Text2D:
    _SHADERS = {
        "vertex": """
    #version 120

    attribute vec2 position;
    attribute vec2 texCoords;
    uniform mat4 matrix;

    varying vec2 vCoords;

    void main(void) {
        gl_Position = matrix * vec4(position, 0.0, 1.0);
        vCoords = texCoords;
    }
    """,
        "fragment": """
    #version 120

    uniform sampler2D texText;
    uniform vec4 color;
    uniform vec4 bgColor;

    varying vec2 vCoords;

    void main(void) {
        if (vCoords.x < 0.0 || vCoords.x > 1.0 || vCoords.y < 0.0 || vCoords.y > 1.0) {
            gl_FragColor = bgColor;
        } else {
            gl_FragColor = mix(bgColor, color, texture2D(texText, vCoords).r);
        }
    }
    """,
    }

    _program = Program(_SHADERS["vertex"], _SHADERS["fragment"], attrib0="position")

    # Discard texture objects when removed from the cache
    _textures = weakref.WeakKeyDictionary()
    """Cache already created textures"""

    def __init__(
        self,
        text: str,
        font: qt.QFont,
        x: float = 0.0,
        y: float = 0.0,
        color: tuple[float, float, float, float] = (0.0, 0.0, 0.0, 1.0),
        bgColor: RGBAColorType | None = None,
        align: str = LEFT,
        valign: str = BASELINE,
        rotate: float = 0.0,
        devicePixelRatio: float = 1.0,
        padding: int = 0,
    ):
        self.devicePixelRatio = devicePixelRatio
        self.font = font
        self._vertices = None
        self._text = text
        self._padding = padding
        self.x = x
        self.y = y
        self.color = color
        self.bgColor = bgColor

        if align not in (LEFT, CENTER, RIGHT):
            raise ValueError("Horizontal alignment not supported: {0}".format(align))
        self._align = align

        if valign not in (TOP, CENTER, BASELINE, BOTTOM):
            raise ValueError("Vertical alignment not supported: {0}".format(valign))
        self._valign = valign

        self._rotate = numpy.radians(rotate)

    def _getTexture(self, dotsPerInch: float) -> tuple[Texture, int]:
        # Retrieve/initialize texture cache for current context
        key = self.text, self.font.key(), dotsPerInch

        context = Context.getCurrent()
        if context not in self._textures:
            self._textures[context] = _Cache(
                callback=lambda key, value: value[0].discard()
            )
        textures = self._textures[context]

        if key not in textures:
            image, offset = font.rasterText(self.text, self.font, dotsPerInch)

            texture = Texture(
                gl.GL_RED,
                data=image,
                minFilter=gl.GL_NEAREST,
                magFilter=gl.GL_NEAREST,
                wrap=(gl.GL_CLAMP_TO_EDGE, gl.GL_CLAMP_TO_EDGE),
            )
            texture.prepare()
            textures[key] = texture, offset

        return textures[key]

    @property
    def text(self) -> str:
        return self._text

    @property
    def padding(self) -> int:
        return self._padding

    def getVertices(self, offset: int, shape: tuple[int, int]) -> numpy.ndarray:
        height, width = shape

        if self._align == LEFT:
            xOrig = 0
        elif self._align == RIGHT:
            xOrig = -width
        else:  # CENTER
            xOrig = -width // 2

        if self._valign == BASELINE:
            yOrig = -offset
        elif self._valign == TOP:
            yOrig = 0
        elif self._valign == BOTTOM:
            yOrig = -height
        else:  # CENTER
            yOrig = -height // 2

        vertices = numpy.array(
            (
                (xOrig, yOrig),
                (xOrig + width, yOrig),
                (xOrig, yOrig + height),
                (xOrig + width, yOrig + height),
            ),
            dtype=numpy.float32,
        )

        cos, sin = numpy.cos(self._rotate), numpy.sin(self._rotate)
        vertices = numpy.ascontiguousarray(
            numpy.transpose(
                numpy.array(
                    (
                        cos * vertices[:, 0] - sin * vertices[:, 1],
                        sin * vertices[:, 0] + cos * vertices[:, 1],
                    ),
                    dtype=numpy.float32,
                )
            )
        )

        return vertices

    def render(self, matrix: numpy.ndarray, dotsPerInch: float):
        if not self.text.strip():
            return

        prog = self._program
        prog.use()

        texUnit = 0
        texture, offset = self._getTexture(dotsPerInch)

        gl.glUniform1i(prog.uniforms["texText"], texUnit)

        mat = numpy.dot(matrix, mat4Translate(int(self.x), int(self.y)))
        gl.glUniformMatrix4fv(
            prog.uniforms["matrix"], 1, gl.GL_TRUE, mat.astype(numpy.float32)
        )

        gl.glUniform4f(prog.uniforms["color"], *self.color)
        if self.bgColor is not None:
            bgColor = self.bgColor
        else:
            bgColor = self.color[0], self.color[1], self.color[2], 0.0
        gl.glUniform4f(prog.uniforms["bgColor"], *bgColor)

        paddingOffset = max(0, int(self.padding * self.devicePixelRatio))
        height, width = texture.shape
        vertices = self.getVertices(
            offset, (height + 2 * paddingOffset, width + 2 * paddingOffset)
        )

        posAttrib = prog.attributes["position"]
        gl.glEnableVertexAttribArray(posAttrib)
        gl.glVertexAttribPointer(posAttrib, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, vertices)

        xoffset = paddingOffset / width
        yoffset = paddingOffset / height
        texCoords = numpy.array(
            (
                (-xoffset, -yoffset),
                (1.0 + xoffset, -yoffset),
                (-xoffset, 1.0 + yoffset),
                (1.0 + xoffset, 1.0 + yoffset),
            ),
            dtype=numpy.float32,
        ).ravel()

        texAttrib = prog.attributes["texCoords"]
        gl.glEnableVertexAttribArray(texAttrib)
        gl.glVertexAttribPointer(texAttrib, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, texCoords)

        with texture:
            gl.glDrawArrays(gl.GL_TRIANGLE_STRIP, 0, 4)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
Demonstration window that displays a wait icon until the plot is updated
"""

import numpy.random
from silx.gui import qt
from silx.gui.widgets.WaitingOverlay import WaitingOverlay
from silx.gui.plot import Plot2D


class MyMainWindow(qt.QMainWindow):
    WAITING_TIME = 2000  # ms

    def __init__(self, parent=None):
        super().__init__(parent)

        # central plot
        self._plot = Plot2D()
        self._waitingOverlay = WaitingOverlay(self._plot)
        self.setCentralWidget(self._plot)

        # button to trigger image generation
        self._rightPanel = qt.QWidget(self)
        self._rightPanel.setLayout(qt.QVBoxLayout())
        self._button = qt.QPushButton("generate image", self)
        self._rightPanel.layout().addWidget(self._button)

        self._dockWidget = qt.QDockWidget()
        self._dockWidget.setWidget(self._rightPanel)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidget)

        # set up
        self._waitingOverlay.hide()
        self._waitingOverlay.setIconSize(qt.QSize(60, 60))
        # connect signal / slot
        self._button.clicked.connect(self._triggerImageCalculation)

    def _generateRandomData(self):
        self.setData(numpy.random.random(1000 * 500).reshape((1000, 500)))
        self._button.setEnabled(True)

    def setData(self, data):
        self._plot.addImage(data)
        self._waitingOverlay.hide()

    def _triggerImageCalculation(self):
        self._plot.clear()
        self._button.setEnabled(False)
        self._waitingOverlay.show()
        qt.QTimer.singleShot(self.WAITING_TIME, self._generateRandomData)


qapp = qt.QApplication([])
window = MyMainWindow()
window.show()
qapp.exec_()



# /*##########################################################################
#
# Copyright (c) 2023 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module provides :class:`PlotWidget`-related QMenu.

The following QMenu is available:

- :class:`ZoomEnabledAxesMenu`
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "12/06/2023"


import weakref
from typing import Optional

from silx.gui import qt

from ..PlotWidget import PlotWidget


class ZoomEnabledAxesMenu(qt.QMenu):
    """Menu to toggle axes for zoom interaction"""

    def __init__(self, plot: PlotWidget, parent: Optional[qt.QWidget] = None):
        super().__init__(parent)
        self.setTitle("Zoom axes")

        assert isinstance(plot, PlotWidget)
        self.__plotRef = weakref.ref(plot)

        self.addSection("Enabled axes")
        self.__xAxisAction = qt.QAction("X axis", parent=self)
        self.__yAxisAction = qt.QAction("Y left axis", parent=self)
        self.__y2AxisAction = qt.QAction("Y right axis", parent=self)

        for action in (self.__xAxisAction, self.__yAxisAction, self.__y2AxisAction):
            action.setCheckable(True)
            action.setChecked(True)
            action.triggered.connect(self._axesActionTriggered)
            self.addAction(action)

        # Listen to interaction configuration change
        plot.interaction().sigChanged.connect(self._interactionChanged)
        # Init the state
        self._interactionChanged()

    def getPlotWidget(self) -> Optional[PlotWidget]:
        return self.__plotRef()

    def _axesActionTriggered(self, checked=False):
        plot = self.getPlotWidget()
        if plot is None:
            return

        plot.interaction().setZoomEnabledAxes(
            self.__xAxisAction.isChecked(),
            self.__yAxisAction.isChecked(),
            self.__y2AxisAction.isChecked(),
        )

    def _interactionChanged(self):
        plot = self.getPlotWidget()
        if plot is None:
            return

        enabledAxes = plot.interaction().getZoomEnabledAxes()
        self.__xAxisAction.setChecked(enabledAxes.xaxis)
        self.__yAxisAction.setChecked(enabledAxes.yaxis)
        self.__y2AxisAction.setChecked(enabledAxes.y2axis)



# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tests for UrlSelectionTable"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "09/05/2023"

import pytest
import weakref
from silx.gui.widgets.UrlSelectionTable import UrlSelectionTable
from silx.gui import qt
from silx.io.url import DataUrl


@pytest.fixture
def urlSelectionTable(qapp, qapp_utils):
    widget = UrlSelectionTable()
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    yield widget
    widget.close()
    ref = weakref.ref(widget)
    widget = None
    qapp_utils.qWaitForDestroy(ref)


def test_show(qapp_utils, urlSelectionTable):
    qapp_utils.qWaitForWindowExposed(urlSelectionTable)


def test_add_urls(urlSelectionTable):
    urlSelectionTable.addUrl(DataUrl("aaaa"))
    urlSelectionTable.addUrl(DataUrl("bbbb"))
    urlSelectionTable.addUrl(DataUrl("cccc"))
    assert urlSelectionTable.rowCount() == 3


def test_clear(urlSelectionTable):
    urlSelectionTable.addUrl(DataUrl("aaaa"))
    assert urlSelectionTable.rowCount() == 1
    urlSelectionTable.clear()
    assert urlSelectionTable.rowCount() == 0


def test_set_remove_error(urlSelectionTable):
    urlSelectionTable.addUrl(DataUrl("aaaa"))
    item = urlSelectionTable._getItemFromUrlPath("aaaa")
    urlSelectionTable.setError("aaaa", "Oh... no...")
    assert not item.icon().isNull()
    urlSelectionTable.setError("aaaa", "")
    assert item.icon().isNull()
